/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.field.model;

import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.io.ParseException;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.geoviewer.Geometrical;
import org.openide.util.NbBundle;
import org.openide.util.NbBundle.Messages;

/**
 * non persistent object for storing geometries and additional information
 *
 *
 * @author Timon Veenstra
 */
public class Shape implements Geometrical<Geometry>, Transferable, Cloneable {

    private static final Logger LOG = Logger.getLogger("nl.cloudfarming.client.field.model.Shape");
    public static final DataFlavor DATA_FLAVOR = new DataFlavor(Shape.class, "Shape");
    private static long lastId = 0l;
    private Geometry geometry;
    private Map<String, String> stringAttributeMap = new HashMap<>();
    private String name;
    private final long id;
    private ShapeType type;
    private ImportType importType = ImportType.IMPORT;
    
    public static final String PROP_IMPORT_TYPE = "importType";
    
    private transient final java.beans.PropertyChangeSupport propertyChangeSupport = new java.beans.PropertyChangeSupport(this);

    /**
     * Add PropertyChangeListener.
     *
     * @param listener
     */
    public void addPropertyChangeListener(java.beans.PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    /**
     * Remove PropertyChangeListener.
     *
     * @param listener
     */
    public void removePropertyChangeListener(java.beans.PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }

    /**
     *
     */
    public Shape() {
        this.id = lastId++;
    }

    /**
     * Construct a shape object based on a WKT string
     *
     * @param Geometry
     * @throws ParseException
     */
    public Shape(Geometry geometry) {
        this.id = lastId++;
        this.geometry = geometry;

        // Determine and set shape type
        final String geomType = geometry.getGeometryType();

        switch (geomType.toUpperCase()) {
            case "MULTIPOLYGON":
                type = ShapeType.FIELD;
                break;
            case "POLYGON":
                type = ShapeType.FIELD;
                break;
            case "LINESTRING":
                type = ShapeType.HELPLINE;
                break;
            case "MULTILINESTRING":
                type = ShapeType.HELPLINE;
                break;
            default:
                type = ShapeType.UNKNOWN;
                break;
        }
    }

    public ShapeType getShapeType() {
        return type;
    }

    public ImportType getImportType() {
        return importType;
    }

    public void setImportType(ImportType importType) {
        ImportType old = this.importType;
        this.importType = importType;
        propertyChangeSupport.firePropertyChange(PROP_IMPORT_TYPE, old, importType);
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    /**
     * @see nl.cloudfarming.client.geoviewer.LayerObject#getGeometry()
     * @return The Geometry of type MultiPolygon
     */
    @Override
    public Geometry getGeometry() {
        return geometry;
    }

    /**
     * @see nl.cloudfarming.client.geoviewer.LayerObject#getTooltipText()
     * @return TooltipText as String
     */
    @Override
    public String getTooltipText() {
        return getName();
    }

    /**
     * @see nl.cloudfarming.client.geoviewer.LayerObject#getId()
     * @return The Id of the wrapped object
     */
    @Override
    public long getId() {
        return id;
    }

    /**
     * Returns the shape type as ShapeType (enum-value) A shape can for example
     * be a Field, or a Line, etc.
     *
     * @return Shape type as ShapeType
     */
    public ShapeType getType() {
        return type;
    }

    /**
     * Set the shape type as ShapeType (enum-value) A shape can for example be a
     * Field, or a Line, etc.
     *
     * @param type Set the type of this Shape
     */
    public void setType(ShapeType type) {
        this.type = type;
    }

    @Override
    public Object clone() throws CloneNotSupportedException {
        return super.clone();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Shape other = (Shape) obj;
        if (this.id != other.id) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        int hash = super.hashCode();
        hash = 79 * hash + (int) (this.id ^ (this.id >>> 32));
        return hash;
    }

    @Override
    public Geometry getRenderObject(Envelope envelope) {
        if (envelope.intersects(geometry.getEnvelopeInternal())) {
            return geometry;
        }
        return null;
    }

    @Override
    public Point getCentroid() {
        return this.geometry.getCentroid();
    }

    @Override
    public DataFlavor[] getTransferDataFlavors() {
        return new DataFlavor[]{DATA_FLAVOR};
    }

    @Override
    public boolean isDataFlavorSupported(DataFlavor flavor) {
        return DATA_FLAVOR.equals(flavor);
    }

    @Override
    public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException {
        if (isDataFlavorSupported(flavor)) {
            return this;
        }
        throw new UnsupportedFlavorException(flavor);
    }

    /**
     * Add a string attribute to this shape
     * a string attribute is additional data that is contained in the geometry file
     * @param name the name of the attribute
     * @param value the value of the attribute
     */
    public void addStringAttribute(String name, String value) {
        stringAttributeMap.put(name, value);
    }
    
    /**
     * Gets a unmodifiable map of the Attributes of this shape
     * This map contains additional information of the shape for instance it could contain the name of the shape
     * @return the map
     */
    public Map<String, String> getStringAttributeMap(){
        return Collections.unmodifiableMap(stringAttributeMap);
    }

    
    /**
     * Set the name of the shape based on a attribute that should be present in the stringAttibute map
     * if the map contains attribute name will set the name of this shape to the value in the map
     * if the parsed attributeName is "" the it will set the name of this shape to ""
     * if the parsed attribute is not found in the map will do nothing and log a warning 
     * @param attributeName the key in the attributeMap
     */
    public void setNameFromAttribute(String attributeName) {
        String value = stringAttributeMap.get(attributeName);

        if (value != null) {
            setName(value);
        } else {
            if ("".equals(attributeName)) {
                setName("");
            } else {
                LOG.log(Level.WARNING, "attempt to setNameFromAttribute with non existing attributeName {0}", attributeName);
            }
        }
    }

    @Messages({
        "ImportType.skip=Skip",
        "ImportType.Import=Import"
    })
    public enum ImportType {

        /**
         * Skip this shape
         */
        SKIP(NbBundle.getMessage(ImportType.class, "ImportType.skip")),
        IMPORT(NbBundle.getMessage(ImportType.class, "ImportType.Import"));
        private String displayName;

        ImportType(String displayName) {
            this.displayName = displayName;
        }

        public String getDisplayName() {
            return displayName;
        }
    }

    /**
     * Type of a shape object Gives an indication about what kind of shape it
     * is, what the purpose is and where it needs to be stored eventually.
     *
     * @author Timon Veenstra
     */
    public enum ShapeType {

        /**
         * Field shape, Geometry will be a MultiPolygon
         */
        FIELD,
        /**
         * ab-line shape, Geometry will be a line
         */
        //        AB_LINE,
        /**
         * helpline shape, Geometry will be a line
         */
        HELPLINE,
        UNKNOWN;
    }
}
